// Ansteuerung eines HD44780 kompatiblen LCD im 4-Bit-Interfacemodus
// http://www.mikrocontroller.net/articles/HD44780
// http://www.mikrocontroller.net/articles/AVR-GCC-Tutorial/LCD-Ansteuerung
//
// Die Pinbelegung ist ber defines in lcd-routines.h einstellbar
// mit ATmega48 ist die Einstellung der clock/8-fuse relevant. Das Programm luft zu schnell.

//Variante fr das invertierend anzeigende 4 x 40 LCD von Pollin fr 13 Euro in 7/2014
//Das LCd hat einen zweiten Enable-Anschlu. Deshalb gibt es hier neu: Enable1 und Enable2.
//Enable1 bedient die oberen zwei Zeilen, Enable2 die unteren beiden

/*
Abgenderte Version zur Ansteuerung eines 4 x 40 LCDs ber den I2C-Porterweiterungsbaustein PCF8574.


- LCD_PORT abndern in eine Variable.

- LCD_EN1
- LCD_EN2
- LCD_RS
- LCD_DB	PC0		hier noch etwas erfinden -->> Wert ist immer 0. Am besten so lassen.

verwendet "pcf8574_set_outputs (pcf8574_adress, LCD_PORT);" um die Variable ber den I2C-Bus zu schicken.

Adresse PCF8574:
0 1 0 0 A2 A1 A0 RW ACK 	-->> 0x40 eins nach rechts schieben -->> 0x20 falls A2, A1, A0 alle GND sind


Adresse PCF8574P:
0 1 1 1 A2 A1 A0 RW ACK 	-->> 0x70 eins nach rechts schieben -->> 0x38 falls A2, A1, A0 alle GND sind
							-->> 0x74 eins nach rechts schieben -->> 0x3A falls A2, A0 alle GND und A1 = + sind <<--
							-->> 0x78 eins nach rechts schieben -->> 0x3C falls A1, A0 alle GND und A2 = + sind 
							-->> 0x72 eins nach rechts schieben -->> 0x39 falls A2, A1 alle GND und A0 = + sind 
*/


#include <avr/io.h>
#include "LCD-RoutinesI2C-4x40.h"	//Fr einen anderen Typ mu hier die umbenannte Datei eingetragen sein
#include <util/delay.h>
//#define pcf8574_adress	0x20	//PCF8574 ohne P alle drei Adress-Pins an GND
//#define pcf8574_adress	0x22	//Pin 2 = A1 des PCF8574 sei an + 5 Volt
#define pcf8574_adress 0x38		//PCF8574P mit P alle drei Adress-Pins an GND


uint8_t LCD_PORT = 0;			//neue Variable, die dann ber I2C an den Porterweiterungsbaustein bergeben wird.
uint8_t oben_oder_unten4x40 = 0;	//entscheidet, ob Enable1 oder Enable2 bettigt werden soll
 
////////////////////////////////////////////////////////////////////////////////
// Erzeugt einen Enable-Puls
//static void I2Clcd_enable( void )		//kann man mit "static" nicht von auerhalb aufrufen
void I2Clcd_enable( void )				//am besten aufteilen in zwei Unterprogramme Enable1 und Enable2 oder Fallunterscheidung machen
{
    if (oben_oder_unten4x40 == 0)	LCD_PORT |= 1<<LCD_EN1;    	 // Enable 1 oder 2  auf 1 setzen
	else LCD_PORT |= 1<<LCD_EN2;
	pcf8574_set_outputs (pcf8574_adress, LCD_PORT);
	
    _delay_us( LCD_ENABLE_US );  // kurze Pause	kann vermutlich entfallen, da der I2C-Buszugriff mindestens genauso lange dauert
    if (oben_oder_unten4x40 == 0)	LCD_PORT &= ~(1<<LCD_EN1);    // Enable 1 oder 2 auf 0 setzen
	else LCD_PORT &= ~(1<<LCD_EN2);
	pcf8574_set_outputs (pcf8574_adress, LCD_PORT);
}

/*
4x40 LCd ausprobiert, geht, zeigt aber in beiden Hlften das gleiche an.
Idee: man knnte gleich beim Aufruf angeben, ob in die obere oder untere Hlfte geschrieben werden soll, anstatt 
anhand der Cursorposition zu entscheiden, ob Enable1 oder Enable2 gegeben werden soll.
*/
 
////////////////////////////////////////////////////////////////////////////////
// Sendet eine 4-bit Ausgabeoperation an das LCD
static void I2Clcd_out( uint8_t data )
{
    data &= 0xF0;                       // obere 4 Bit maskieren
 
    LCD_PORT &= ~(0xF0>>(4-LCD_DB));    // Maske lschen
	pcf8574_set_outputs (pcf8574_adress, LCD_PORT);
	
    LCD_PORT |= (data>>(4-LCD_DB));     // Bits setzen
	pcf8574_set_outputs (pcf8574_adress, LCD_PORT);
	I2Clcd_enable();
//    if (oben_oder_unten == 0) I2Clcd_enable(0);
//	else I2Clcd_enable(1);
}


void I2Clcd_init2( void )	//war vorher Bestandteil von "I2Clcd_init()"
{
    // warten auf die Bereitschaft des LCD
    _delay_ms( LCD_BOOTUP_MS );
    
    // Soft-Reset muss 3mal hintereinander gesendet werden zur Initialisierung
    I2Clcd_out( LCD_SOFT_RESET );
    _delay_ms( LCD_SOFT_RESET_MS1 );
 
    I2Clcd_enable();
    _delay_ms( LCD_SOFT_RESET_MS2 );
 
    I2Clcd_enable();
    _delay_ms( LCD_SOFT_RESET_MS3 );
 
    // 4-bit Modus aktivieren 
    I2Clcd_out( LCD_SET_FUNCTION |
             LCD_FUNCTION_4BIT );
    _delay_ms( LCD_SET_4BITMODE_MS );
 
    // 4-bit Modus / 2 Zeilen / 5x7
    I2Clcd_command( LCD_SET_FUNCTION |
                 LCD_FUNCTION_4BIT |
                 LCD_FUNCTION_2LINE |
                 LCD_FUNCTION_5X7 );
 
    // Display ein / Cursor aus / Blinken aus
    I2Clcd_command( LCD_SET_DISPLAY |
                 LCD_DISPLAY_ON |
                 LCD_CURSOR_OFF |
                 LCD_BLINKING_OFF); 
 
    // Cursor inkrement / kein Scrollen
    I2Clcd_command( LCD_SET_ENTRY |
                 LCD_ENTRY_INCREASE |
                 LCD_ENTRY_NOSHIFT );
 
    I2Clcd_clear();

}
 
////////////////////////////////////////////////////////////////////////////////
// Initialisierung: muss ganz am Anfang des Programms aufgerufen werden.
void I2Clcd_init( void )
{
/*    // verwendete Pins auf Ausgang schalten
	uint8_t pins = (0x0F << LCD_DB) |          // 4 Datenleitungen
                   (1<<LCD_RS) |               // R/S Leitung
                   (1<<LCD_EN1 ;				// Enable Leitung
*/	
	
    uint8_t pins = (0x0F << LCD_DB) |          // 4 Datenleitungen
                   (1<<LCD_RS) |               // R/S Leitung
                   (1<<LCD_EN1) |				// Enable Leitung
				   (1<<LCD_EN2) |
				   (1<<LCD_RW);                // RW-Leitung soll immer low sein, so lange man nicht lesen mchte
//    LCD_DDR |= pins;							//fllt bei I2C ganz weg
 
    // initial alle Ausgnge auf Null
    LCD_PORT &= ~pins;
	pcf8574_set_outputs (pcf8574_adress, LCD_PORT);
	
	I2Clcd_init2();	
	oben_oder_unten4x40 = 1;			//umschalten auf die andere Hlfte
	I2Clcd_init2();
 
/*    // warten auf die Bereitschaft des LCD
    _delay_ms( LCD_BOOTUP_MS );
    
    // Soft-Reset muss 3mal hintereinander gesendet werden zur Initialisierung
    I2Clcd_out( LCD_SOFT_RESET );	//obere Hlfte
	oben_oder_unten = 1;			//umschalten auf die andere Hlfte
	I2Clcd_out( LCD_SOFT_RESET );	//untere Hlfte
    _delay_ms( LCD_SOFT_RESET_MS1 );
 
    I2Clcd_enable();
    _delay_ms( LCD_SOFT_RESET_MS2 );
 
    I2Clcd_enable();
    _delay_ms( LCD_SOFT_RESET_MS3 );
 
    // 4-bit Modus aktivieren 
	oben_oder_unten = 0;			//umschalten auf die andere Hlfte
    I2Clcd_out( LCD_SET_FUNCTION |
             LCD_FUNCTION_4BIT);	//obere Hlfte
	oben_oder_unten = 1;			//umschalten auf die andere Hlfte
    I2Clcd_out( LCD_SET_FUNCTION |
             LCD_FUNCTION_4BIT);	//untere Hlfte
    _delay_ms( LCD_SET_4BITMODE_MS );
 
    // 4-bit Modus / 2 Zeilen / 5x7
    I2Clcd_command( LCD_SET_FUNCTION |
                 LCD_FUNCTION_4BIT |
                 LCD_FUNCTION_2LINE |
                 LCD_FUNCTION_5X7 );
 
    // Display ein / Cursor aus / Blinken aus
    I2Clcd_command( LCD_SET_DISPLAY |
                 LCD_DISPLAY_ON |
                 LCD_CURSOR_OFF |
                 LCD_BLINKING_OFF); 
 
    // Cursor inkrement / kein Scrollen
    I2Clcd_command( LCD_SET_ENTRY |
                 LCD_ENTRY_INCREASE |
                 LCD_ENTRY_NOSHIFT );
 
    I2Clcd_clear();*/
}
  
////////////////////////////////////////////////////////////////////////////////
// Sendet ein Datenbyte an das LCD
void I2Clcd_data( uint8_t data )
{
    LCD_PORT |= (1<<LCD_RS);    // RS auf 1 setzen
	pcf8574_set_outputs (pcf8574_adress, LCD_PORT);
 
    I2Clcd_out( data );            // zuerst die oberen, 
    I2Clcd_out( data<<4 );         // dann die unteren 4 Bit senden
 
    _delay_us( LCD_WRITEDATA_US );
}
 
////////////////////////////////////////////////////////////////////////////////
// Sendet einen Befehl an das LCD
//"oben_oder_unten" durchreichen weiter zu "I2Clcd_out".
void I2Clcd_command( uint8_t data )
{
    LCD_PORT &= ~(1<<LCD_RS);    // RS auf 0 setzen
	pcf8574_set_outputs (pcf8574_adress, LCD_PORT);
 
    I2Clcd_out( data );             // zuerst die oberen, 
    I2Clcd_out( data<<4 );           // dann die unteren 4 Bit senden
 
    _delay_us( LCD_COMMAND_US );
}
 
////////////////////////////////////////////////////////////////////////////////
// Sendet den Befehl zur Lschung des Displays	-->> Fehler!!! Lscht nur die obere Hlfte!!
/*
void I2Clcd_clear( void )
{
    I2Clcd_command( LCD_CLEAR_DISPLAY );
    _delay_ms( LCD_CLEAR_DISPLAY_MS );
}
*/
void I2Clcd_clear( void )
{
	oben_oder_unten4x40 = 0;
    I2Clcd_command( LCD_CLEAR_DISPLAY );
    _delay_ms( LCD_CLEAR_DISPLAY_MS );
	
	oben_oder_unten4x40 = 1;	
	I2Clcd_command( LCD_CLEAR_DISPLAY );
    _delay_ms( LCD_CLEAR_DISPLAY_MS );
}
 
////////////////////////////////////////////////////////////////////////////////
// Sendet den Befehl: Cursor Home
void I2Clcd_home( void )
{
    I2Clcd_command( LCD_CURSOR_HOME );
    _delay_ms( LCD_CURSOR_HOME_MS );
}
 
////////////////////////////////////////////////////////////////////////////////
// Setzt den Cursor in Spalte x (0..39) Zeile y (1..4) 
//neue Variable "oben_oder_unten" hinzu gefgt, um Enable1 oder Enable2 bettigen. Sie wird durch "I2Clcd_command" durchgereicht bis zu "I2Clcd_out".
//x geht von 0..39 und y von 1...4 
void I2Clcd_setcursor( uint8_t x, uint8_t y )
{
    uint8_t data;
 
//	if ( x >= 40 ) return;		//Funktion abbrechen und nichts machen, falls zu hoher x-Wert bergeben wurde. -->> ungnstig
//	if ( x >= 40 ) x = 0;		// an erste Stelle zurck springen, falls zu hoher x-Wert bergeben wurde.
	if ( x >= 40 ) x = 39;		// an letzter Stelle verharren, falls zu hoher x-Wert bergeben wurde.

    switch (y)
    {
        case 1:    // 1. Zeile  -->> per Paramerter bestimmen, da "I2Clcd_command" das Enable1 bettigt.
            data = LCD_SET_DDADR + LCD_DDADR_LINE1 + x;
			oben_oder_unten4x40 = 0;
            break;
 
        case 2:    // 2. Zeile  -->> per Paramerter bestimmen, da "I2Clcd_command" das Enable1 bettigt.
            data = LCD_SET_DDADR + LCD_DDADR_LINE2 + x;
			oben_oder_unten4x40 = 0;
            break;
 
        case 3:    // 3. Zeile	-->> per Paramerter bestimmen, da "I2Clcd_command" das Enable2 bettigt.
            data = LCD_SET_DDADR + LCD_DDADR_LINE3 + x;
			oben_oder_unten4x40 = 1;
            break;
 
        case 4:    // 4. Zeile  -->> per Paramerter bestimmen, da "I2Clcd_command" das Enable2 bettigt.
            data = LCD_SET_DDADR + LCD_DDADR_LINE4 + x;
			oben_oder_unten4x40 = 1;
            break;
 
        default:
            return;                                   // fr den Fall einer falschen Zeile
    }

//hier mte entschieden werden anhand des Wertes in "data", ob Enable1 oder Enable2 gegeben werden soll
//ab hier alle nachfolgenden Aufrufe von "LCD_command" mten bestimmen knnen, ob das nachfolgende "LCD_out" das Enable1 oder Enable2 bedient.
//-->> man bentigt einen zustzlichen Parameter, LCD_out doppelt anlegen oder es mu per Parameter eine Fallunterscheidung machen knnen.
//am besten vonhier aus eine globale Variable verndern, die dann in "lcd_out" abgefragt wird.
    I2Clcd_command( data );
//    I2Clcd_command( data, oben_oder_unten );

}
 
////////////////////////////////////////////////////////////////////////////////
// Schreibt einen String auf das LCD
 
void I2Clcd_string( const char *data )
{
    while( *data != '\0' )
        I2Clcd_data( *data++ );
}

////////////////////////////////////////////////////////////////////////////////
// Schreibt einen String auf das LCD, wurde zur Ausgabe beim RFM69 verwendet
 
//void I2Clcd_string_2( const char *data )	//zum Lesen aus einem Array, das im Flash definiert ist
void I2Clcd_string_2( char *data )			//zum LEsen aus einem variablen Array im RAM
{
char q;
//  while( *data <= SPIRead(0x38) )
//	while( q <= 38 )
	while( (q <= 38) && ( *data != '\0' ) )
		{
        I2Clcd_data( *data++ );
		q++;
		_delay_ms(60);
		}
	//Schrift erscheint kurz, wird aber sofort wieder von den unerwnschten Zeichen berschrieben. SCheint stndig zu schreiben.
	//prfen, welches Zeichen im String berall steht, mte 0x00 sein. Falls 0x00 steht, nicht schreiben.	
	//q scheint bei jedem Eintritt ins Unterprogramm wieder 0 zu sein. Mte man mal direkt als Hex ausgeben lassen.
}

 //Register Nr 0x38 lesen, und nur soundsooft das nchste Stringelement lesen, wie aus Reg0x38 gelesen wurde.
 //0x3825,			//37 Zeichen
 //man kann auch "(RFM69ConfigTbl[15] & 0xFF)" einsetzen -->> ausprobieren.
 
////////////////////////////////////////////////////////////////////////////////
// Schreibt ein Zeichen in den Character Generator RAM

/* 
void I2Clcd_generatechar( uint8_t code, const uint8_t *data )
{
    // Startposition des Zeichens einstellen
    I2Clcd_command( LCD_SET_CGADR | (code<<3) );
 
    // Bitmuster bertragen
    for ( uint8_t i=0; i<8; i++ )
    {
        I2Clcd_data( data[i] );
    }
}
*/


//-------------------------------------------------------------------------------------------------
// LCD_WriteHex schreibt einByte als Hex-Zahl aufs Display
//-------------------------------------------------------------------------------------------------
inline unsigned char nibble_zu_hex(unsigned char nib) {
 if (nib < 0x0A) nib = nib + '0';
 else nib = (nib - 0x0A) + 'A';
 return nib;
}

void I2CLCD_WriteHex(unsigned char zahl)
{
	unsigned char temp; 
	temp = (zahl & 0x0F); 
	zahl = (zahl>>4); 
	I2Clcd_data(nibble_zu_hex(zahl)); //das erste Zeichen aufs LCD ausgeben
	I2Clcd_data(nibble_zu_hex(temp)); //das zweite Zeichen aufs LCD ausgeben
}

