#include "pcf8574.h"
#include "LCD-RoutinesI2C-4x40.h"	//fr neue Schaltung


/*
"Note: TWBR should be 10 or higher if the TWI operates in Master mode.
If TWBR is lower than 10, the Master may produce an incorrect output on SDA and SCL for the
reminder of the byte. The problem occurs when operating the TWI in Master mode, sending
Start + SLA + R/W to a Slave (a Slave does not need to be connected to the bus for the condition to happen)."
*/

void pcf8574_init (void)
{
	/*set bus speed*/	//TWBR soll nicht kleiner als 10 sein!!!
//	TWBR = 0x10;	//	166 KHz	OK, geht nicht mit 18 kOhm
	TWBR = 0x40;	//	55,5 kHz, luft nur ca 10 Sekunden lang mit 18 kOhm und bleibt dann mit Fehler stehen
//	TWBR = 0x50;	//	45,5 kHz, geht noch mit 18 kOhm - Widerstnden am Bus, bleibt nach 150 Sekunden mit Fehler stehen
//	TWBR = 0x80;	// -->> 29 kHz, geht noch mit 18 kOhm - Widerstnden am Bus
	//fSCK = 8 MHz / (16 + 2 TWBR)	-->> 8000000 / (16 + 2 * 128) = 29411 Hz
	
	
	//ab hier Versuche mit zwei 1,8 k Widerstnden am I2C-bus:
//	TWBR = 0x10;	//		166 KHz	OK, ist schnell und Bus luft stabil, aber RC5 reagiert fast nicht mehr.
//	TWBR = 0x04;	//		333 kHz
//	TWBR = 0x03;	//		364 kHz
//	TWBR = 0x02;	//		400 kHz	OK, ist schnell und Bus luft stabil, aber RC5 reagiert fast nicht mehr.

	TWCR= 0x04; //enable twi	The TWEN bit enables TWI operation and activates the TWI interface.
}


unsigned char pcf8574_send_start (void)
{
	/*writing a one to TWINT clears it, TWSTA=Start, TWEN=TWI-enable*/
	TWCR = (1<<TWINT) | (1<<TWSTA) | (1<<TWEN);
	/*wait, until start condition has been sent --> ACK*/
	while (!(TWCR & (1<<TWINT)));

	#ifdef debug_GLCD
	GLCD_WriteHex ( 4,0, TWSR & 0b11111100);		//-->> 0x08 falls I2C geht
	#endif
//	LCD_Ausgabe (TWSR);
//	I2Clcd_setcursor( 13, 1 );		//geht nicht, ruft die Funktion immer wieder unendlich oft auf.
//	I2CLCD_WriteHex(TWSR & 0b11111100);

	return TWSR;
}


void pcf8574_send_stop (void)
{
	/*writing a one to TWINT clears it, TWSTO=Stop, TWEN=TWI-enable*/
	TWCR = (1<<TWINT) | (1<<TWSTO) | (1<<TWEN);	
	#ifdef debug_GLCD
	GLCD_WriteHex ( 6,0, TWSR & 0b11111100);	//-->> 0x28 falls I2C geht
	#endif
}


unsigned char pcf8574_send_add_rw (unsigned char address, unsigned char rw)
{
	/*address can be 0 .. 8; rw=0 --> write, rw=1 --> read*/
	unsigned char addr_byte = 0;
	/*shift address one bit left*/
	addr_byte = address << 1;
	/*set RW-Bit, if necessary*/
	addr_byte |= rw;
	/*0b0100xxx0 --> address of Expander*/
	addr_byte |= 0b01000000;
	/*TWDR contains byte to send*/
	TWDR = addr_byte;
	/*send content of TWDR*/
	TWCR = (1<<TWINT) | (1<<TWEN);
	/*wait, until address has been sent --> ACK*/
	while (!(TWCR & (1<<TWINT)));
	#ifdef debug_GLCD
	GLCD_WriteHex ( 4,1, TWSR & 0b11111100);	//-->> 0x18 falls I2C geht
	#endif	
	return TWSR;
}


unsigned char pcf8574_send_byte (unsigned char byte)
{
	/*TWDR contains byte to send*/
	TWDR = byte;
	/*send content of TWDR*/
	TWCR = (1<<TWINT) | (1<<TWEN);
	/*wait, until byte has been sent --> ACK*/
	while (!(TWCR & (1<<TWINT)));
	#ifdef debug_GLCD
	GLCD_WriteHex ( 6,1, TWSR & 0b11111100);	//-->> 0x28 falls I2C geht
	#endif	
	return TWSR;
}


unsigned char pcf8574_read_byte (void)
{
	/*send content of TWDR; TWEA = enable ACK*/
	TWCR = (1<<TWINT) | (1<<TWEA) | (1<<TWEN);
	/*wait, until byte has been received --> ACK*/
	while (!(TWCR & (1<<TWINT)));
	return TWDR;
}

/*
unsigned char pcf8574_get_inputs (unsigned char address)
{
	pcf8574_init ();
	pcf8574_send_start ();
	pcf8574_send_add_rw (address, 1);			//Adresse senden und Lesen einstellen
	unsigned char input = pcf8574_read_byte ();
	pcf8574_send_stop ();
	return input;
}
*/

void pcf8574_set_outputs (unsigned char address, unsigned char byte)
{
//	pcf8574_init ();
	pcf8574_send_start ();
	pcf8574_send_add_rw (address, 0);			//Adresse senden und Schreiben einstellen
	pcf8574_send_byte (byte);
	pcf8574_send_stop ();
}

